#include <assert.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lib/include/ns_fget_line.h"
#include "lib/include/ns_str.h"

double read_file(const char *filename) {
  FILE *fp;
  int status;
  Str line = {0};
  char *pss;

  double sum = 0;
  size_t line_pss = 0;

  errno = 0;
  fp = fopen(filename, "r");

  if (fp == NULL) {
    return 0;
  }

  while (1) {
    ns_str_init_empty(&line, 256);

    errno = 0;
    status = ns_fget_line(fp, &line);

    if (status == EOF) {
      if (errno != 0) {
        perror("error reading file");
        ns_str_destroy(&line);
        return 0;
      } else {
        ns_str_destroy(&line);
        break;
      }
    }

    status = sscanf(line.arr, "Pss: %lu", &line_pss);

    if (status != 1) {
      ns_str_destroy(&line);
      continue;
    }

    sum += line_pss;

    ns_str_destroy(&line);
  }

  fclose(fp);

  return sum;
}

int main(int argc, char **argv) {
  DIR *proc_dp;
  struct dirent *proc_ent;
  Str smaps_path = {0};
  double total_KiB = 0;

  errno = 0;
  proc_dp = opendir("/proc");

  if (proc_dp == NULL) {
    perror("error opening /proc for enumeration");
    assert(proc_dp != NULL);
  }

  while (1) {
    errno = 0;
    proc_ent = readdir(proc_dp);

    if (proc_ent == NULL) {
      if (errno != 0) {
        perror("error enumerating /proc");
        assert(errno == 0);
      } else {
        break;
      }
    }

    // skip if not a PID
    if (!isdigit(proc_ent->d_name[0])) continue;

    ns_str_set(&smaps_path, "/proc/");
    ns_str_push(&smaps_path, proc_ent->d_name);
    ns_str_push(&smaps_path, "/smaps");

    total_KiB += read_file(smaps_path.arr);

    ns_str_destroy(&smaps_path);
  }

  closedir(proc_dp);

  if (argc == 2) {
    if ((strcmp(argv[1], "--pretty") == 0) || (strcmp(argv[1], "-p") == 0)) {
      const char *prefixes[] = {"KiB", "MiB", "GiB"};
      size_t i = 0;
      while (total_KiB >= 1024 && i < 2) {
        total_KiB /= 1024;
        i++;
      }

      printf("%0.2f %s\n", total_KiB, prefixes[i]);
      exit(0);
    } else if ((strcmp(argv[1], "--help") == 0) ||
               (strcmp(argv[1], "-h") == 0)) {
      printf("DESCRIPTION\n");
      printf(
          "  pss-total: sum proportional set size of every process for "
          "accurate memory\n");
      printf(
          "  usage statistics. In non-pretty mode, the number is given in "
          "kibibytes.\n");
      printf("\n");
      printf("SUMMARY\n");
      printf("  pss-total [OPTIONS]\n");
      printf("\n");
      printf("OPTIONS\n");
      printf("  -p, --pretty\n");
      printf("      format output using sensible units: KiB, MiB, or GiB.\n");
      printf("\n");
      printf("  -h, --help\n");
      printf("      print this help message.\n");
      exit(0);
    }
  }

  printf("%.0f\n", total_KiB);
  return 0;
}
